import EMDY from "@endeavorance/emdy";
import TOML from "smol-toml";
import YAML from "yaml";
import type { LoadedFile } from "./files";

export type UnknownRecord = Record<string, unknown>;

/**
 * Given an unknown shape, ensure it is an object or array of objects,
 * then return it as an array of objects.
 *
 * @param val - The unknown value to format
 * @returns - An array of objects
 */
function formatEntries(val: unknown): UnknownRecord[] {
  if (Array.isArray(val) && val.every((el) => typeof el === "object")) {
    return val;
  }

  if (typeof val === "object" && val !== null) {
    return [val as UnknownRecord];
  }

  throw new Error(
    `Invalid data format. Entry files must define an object or array of objects, but found "${val}"`,
  );
}

/**
 * Parse one or more YAML documents from a string
 *
 * @param text - The YAML string to parse
 * @returns - An array of parsed objects
 */
export function parseYAMLEntries(text: string): UnknownRecord[] {
  const parsedDocs = YAML.parseAllDocuments(text);

  if (parsedDocs.some((doc) => doc.toJS() === null)) {
    throw new Error("Encountered NULL resource");
  }

  const errors = parsedDocs.flatMap((doc) => doc.errors);

  if (errors.length > 0) {
    throw new Error(
      `Error parsing YAML resource: ${errors.map((e) => e.message).join(", ")}`,
    );
  }

  const collection: UnknownRecord[] = parsedDocs.map((doc) => doc.toJS());
  return collection;
}

/**
 * Parse one or more JSON documents from a string
 *
 * @param text - The JSON string to parse
 * @returns - An array of parsed objects
 */
export function parseJSONEntries(text: string): UnknownRecord[] {
  return formatEntries(JSON.parse(text));
}

/**
 * Parse one or more TOML documents from a string
 *
 * @param text - The TOML string to parse
 * @returns - An array of parsed objects
 */
export function parseTOMLEntries(text: string): UnknownRecord[] {
  return formatEntries(TOML.parse(text));
}

/**
 * Parse one or more Markdown documents from a string
 *
 * @param text - The Markdown string to parse
 * @returns - An array of parsed objects
 */
export function parseMarkdownEntry(
  text: string,
  contentKey: string,
): UnknownRecord[] {
  return formatEntries(EMDY.parse(text, contentKey));
}

export function autoParseEntry(loadedFile: LoadedFile, contentKey: string) {
  const { fileType, content } = loadedFile;

  if (fileType === "md") {
    return parseMarkdownEntry(content, contentKey);
  }

  if (fileType === "yaml") {
    return parseYAMLEntries(content);
  }

  if (fileType === "json") {
    return parseJSONEntries(content);
  }

  if (fileType === "toml") {
    return parseTOMLEntries(content);
  }

  throw new Error(`Unsupported file type: ${fileType}`);
}
